#! /usr/bin/env python

import logging

from .. import iso8601 as iso
from .. import rfc2396 as uri

from ..py2 import (
    byte,
    dict_items,
    force_text,
    is_string,
    is_text,
    is_unicode,
    py2,
    range3,
    SortableMixin)
from ..unicode5 import BasicParser

from .grammar import WordParser, COMMA, is_digits, quote_string
from .grammar import format_parameters


class Parameter(object):

    """Abstract base class for HTTP Parameters

    Provides conversion to strings based on the :meth:`to_bytes` method.
    In Python 2, also provides conversion to the unicode string type.
    In Python 3, implements __bytes__ to enable use of bytes(parameter)
    which becomes portable as in Python 2 __str__ is mapped to to_bytes
    too.

    The HTTP grammar and the parsers and classes that implement it all
    use binary strings but usage of byte values outside those of the US
    ASCII codepoints is discouraged and unlikely to be portable between
    systems.

    When required, Pyslet converts to character strings using the
    ISO-8859-1 codec.  This ensures that the conversions never generate
    unicode decoding erros and is consistent with the text of RFC2616.

    As the purpose of these modules is to provide a way to use HTTP
    constructs in other contexts too, parameters use *character* strings
    where possible.  Therefore, if an attribute must represent a token
    then it is converted to a character string and *must* therefore be
    compared using character strings and not binary strings.  For
    example, see the type and subtype attributes of :class:`MediaType`.
    Similarly where tokens are passed as arguments to constructors these
    must also be character strings.

    Where an attribute may be, or may contain, a value that would be
    represented as a quoted string in the protocol then it is stored as
    a binary string.  You need to take particular care with parameter
    lists as the parameter names are tokens so are character strings but
    the parameter values are binary strings.  The distinction is lost in
    Python 2 but the following code snippet will behave unexpectedly in
    Python 3 so for future compatibility it is better to make usage
    explicit now::

        Python 2.7
        >>> from pyslet.http.params import MediaType
        >>> t = MediaType.from_str("text/plain; charset=utf-8")
        >>> "Yes" if t["charset"] == 'utf-8' else "No"
        'Yes'
        >>> "Yes" if t["charset"] == b'utf-8' else "No"
        'Yes'

        Python 3.5
        >>> from pyslet.http.params import MediaType
        >>> t = MediaType.from_str("text/plain; charset=utf-8")
        >>> "Yes" if t["charset"] == 'utf-8' else "No"
        'No'
        >>> "Yes" if t["charset"] == b'utf-8' else "No"
        'Yes'

    Such values *may* be set using character strings, in which case
    ISO-8859-1 is used to encode them."""

    @classmethod
    def bstr(cls, arg):
        """Returns arg as a binary string"""
        if is_unicode(arg):
            return arg.encode('latin1')
        else:
            return arg

    @classmethod
    def bparameters(cls, parameters):
        """Ensures parameter values are binary strings"""
        result = {}
        warn = False
        for k, v in dict_items(parameters):
            if not is_text(k):
                warn = True
                k = k.decode('ascii')
            n, v = v
            if not is_text(n):
                warn = True
                n = n.decode('ascii')
            if is_unicode(v):
                v = v.encode('iso-8859-1')
            result[k] = (n, v)
        if warn:
            logging.warn("Parameter names should be character strings: %s" %
                         repr(parameters))
        return result

    def to_bytes(self):
        """Returns a binary string representation of the parameter

        This method should be used in preference to str for
        compatibility with Python 3."""
        raise NotImplementedError

    if py2:
        def __str__(self):  # noqa
            return self.to_bytes()

        def __unicode__(self):  # noqa
            return self.to_bytes().decode('iso-8859-1')
    else:
        def __str__(self):  # noqa
            return self.to_bytes().decode('iso-8859-1')

        def __bytes__(self):    # noqa
            return self.to_bytes()


class SortableParameter(SortableMixin, Parameter):

    """Base class for sortable parameters

    Inherits from :class:`~pyslet.py2.SortableMixin` allowing sorting to
    be implemented using a class-specific sortkey method implementation.

    A __hash__ implementation that calls sortkey is also provided to
    enable instances to be used as dictionary keys."""

    def otherkey(self, other):
        """Overridden to provide comparison with strings.

        If other is of either character or binary string types then
        it is passed to the classmethod from_str which is assumed
        to return a new instance of the same class as *self* which
        can then be compared by the return value of sortkey.

        This enables comparisons such as the following::

            >>> t = MediaType.from_str("text/plain")
            >>> t == "text/plain"
            True
            >>> t > "image/png"
            True
            >>> t < "video/mp4"
            True
        """
        if isinstance(other, self.__class__):
            return other.sortkey()
        elif is_string(other):
            return self.from_str(other).sortkey()
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.sortkey())


class HTTPVersion(SortableParameter):

    """Represents the HTTP Version.

    major
        The (optional) major version as an int

    minor
        The (optional) minor version as an int

    The default instance, HTTPVersion(), represents HTTP/1.1

    HTTPVersion objects are sortable (such that 1.1 > 1.0 and  1.2 <
    1.25).

    On conversion to a string the output is of the form::

            HTTP/<major>.<minor>

    For convenience, the constants HTTP_1p1 and HTTP_1p0 are provided
    for comparisons, e.g.::

        if HTTPVersion.from_str(version_str) < HTTP_1p1:
            # do something to support a legacy system..."""

    def __init__(self, major=1, minor=None):
        #: major protocol version (read only)
        self.major = major
        #: minor protocol version (read only)
        self.minor = minor
        if minor is None:
            self.minor = 1 if major == 1 else 0

    @classmethod
    def from_str(cls, source):
        """Constructs an :py:class:`HTTPVersion` object from a string."""
        wp = ParameterParser(source)
        result = wp.require_http_version()
        wp.require_end("HTTP version")
        return result

    def to_bytes(self):
        return b"HTTP/%i.%i" % (self.major, self.minor)

    def sortkey(self):
        return (self.major, self.minor)


#: A constant representing HTTP/1.1
HTTP_1p1 = HTTPVersion(1, 1)

#: A constant representing HTTP/1.0
HTTP_1p0 = HTTPVersion(1, 0)


class HTTPURL(uri.ServerBasedURL):

    """Represents http URLs"""

    #: the default HTTP port
    DEFAULT_PORT = 80

    def __init__(self, octets=b'http://localhost/'):
        super(HTTPURL, self).__init__(octets)

    def canonicalize(self):
        """Returns a canonical form of this URI

        This method is almost identical to the implementation in
        :class:`~pyslet.rfc2396.ServerBasedURL` except that a
        missing path is replaced by '/' in keeping with rules for making
        HTTP requests."""
        new_uri = []
        if self.scheme is not None:
            new_uri.append(self.scheme.lower())
            new_uri.append(':')
        if self.authority is not None:
            new_uri.append('//')
            if self.userinfo is not None:
                new_uri.append(self.userinfo)
                new_uri.append('@')
            new_uri.append(self.host.lower())
            if self.port:  # port could be an empty string
                port = int(self.port)
                if port != self.DEFAULT_PORT:
                    new_uri.append(':')
                    new_uri.append("%i" % int(self.port))
        if self.abs_path is not None:
            if not self.abs_path:
                new_uri.append("/")
            else:
                new_uri.append(uri.canonicalize_data(self.abs_path))
        elif self.rel_path is not None:
            new_uri.append(uri.canonicalize_data(self.rel_path))
        if self.query is not None:
            new_uri.append('?')
            new_uri.append(uri.canonicalize_data(self.query))
        if self.fragment is not None:
            new_uri.append('#')
            new_uri.append(self.fragment)
        return uri.URI.from_octets(''.join(new_uri))


uri.URI.register('http', HTTPURL)


class HTTPSURL(HTTPURL):

    """Represents https URLs"""

    #: the default HTTPS port
    DEFAULT_PORT = 443

    def __init__(self, octets=b'https://localhost/'):
        super(HTTPSURL, self).__init__(octets)


uri.URI.register('https', HTTPSURL)


class FullDate(Parameter, iso.TimePoint):

    """A special sub-class for HTTP-formatted dates

    We extend the basic ISO :class:`~pyslet.iso8601.TimePoint`,
    mixing in the :class:`Parameter` base class and providing an
    implementation of to_bytes.

    The effect is to change the way instances are formatted while
    retaining other timepoint features, including comparisons. Take care
    not to pass an instance as an argument where a plain TimePoint is
    expected as unexpected formatting errors could result.  You can
    always wrap an instance to convert between the two types::

        >>> from pyslet.iso8601 import TimePoint
        >>> from pyslet.http.params import FullDate
        >>> eagle = TimePoint.from_str('1969-07-20T15:17:40-05:00')
        >>> print eagle
        1969-07-20T15:17:40-05:00
        >>> eagle = FullDate(eagle)
        >>> print eagle
        Sun, 20 Jul 1969 20:17:40 GMT
        >>> eagle = TimePoint(eagle)
        >>> print eagle
        1969-07-20T15:17:40-05:00

    Notice that when formatting the date is *always* expressed in GMT as
    per the recommendation in the HTTP specification."""

    @classmethod
    def from_http_str(cls, source):
        """Returns an instance parsed from an HTTP formatted string

        There are three supported formats as described in the
        specification::

            "Sun, 06 Nov 1994 08:49:37 GMT"
            "Sunday, 06-Nov-94 08:49:37 GMT"
            "Sun Nov  6 08:49:37 1994"
        """
        wp = ParameterParser(source)
        tp = wp.require_fulldate()
        wp.parse_sp()
        wp.require_end("full date")
        return tp

    def to_bytes(self):
        """Formats the instance according to RFC 1123

        The format is as follows::

            Sun, 06 Nov 1994 08:49:37 GMT

        This format is also described in in RFC2616 in the production
        rfc1123-date."""
        z = self.shift_zone(0)
        (century, year, month, day,
         hour, minute, second) = z.get_calendar_time_point()
        century, decade, dyear, week, dayofweek = z.date.get_week_day()
        return b"%s, %02i %s %04i %02i:%02i:%02i GMT" % (
            ParameterParser.wkday[dayofweek - 1],
            day,
            ParameterParser.month[month - 1],
            century * 100 + year,
            hour, minute, second)


class TransferEncoding(SortableParameter):

    """Represents an HTTP transfer-encoding.

    token
        The transfer encoding identifier, defaults to "chunked"

    parameters
        A parameter dictionary mapping parameter names to tuples
        of strings: (parameter name, parameter value)

    When sorted, the order in which parameters were parsed is ignored.
    Instances are supported first by token and then by alphabetical
    parameter name/value pairs."""

    def __init__(self, token="chunked", parameters={}):
        token = token.lower()
        if token == "chunked":
            #: the lower-cased transfer-encoding token (defaults to "chunked")
            self.token = "chunked"
            #: declared extension parameters
            self.parameters = {}
        else:
            self.token = token
            if parameters:
                self.parameters = self.bparameters(parameters)
            else:
                self.parameters = {}
        self._hp = sorted((p[0], p[1][1]) for p in dict_items(self.parameters))
        self._hp = tuple(self._hp)

    @classmethod
    def from_str(cls, source):
        """Parses the transfer-encoding from a *source* string.

        If the encoding is not parsed correctly BadSyntax is raised."""
        p = ParameterParser(source)
        te = p.require_transfer_encoding()
        p.require_end("transfer-encoding")
        return te

    @classmethod
    def list_from_str(cls, source):
        """Creates a list of transfer-encodings from a string

        Transfer-encodings are comma-separated"""
        telist = []
        p = ParameterParser(source)
        p.parse_sp()
        while p.the_word is not None:
            if p.parse_separator(COMMA):
                continue
            else:
                te = p.require_transfer_encoding()
                telist.append(te)
        return telist

    def to_bytes(self):
        return self.token.encode('ascii') + format_parameters(self.parameters)

    def sortkey(self):
        return (self.token, self._hp)


class Chunk(SortableParameter):

    """Represents an HTTP chunk header

    size
        The size of this chunk (defaults to 0)

    extensions
        A parameter dictionary mapping parameter names to tuples
        of strings: (chunk-ext-name, chunk-ext-val)

    For completeness, instances are sortable by size and then by
    alphabetical parameter name, value pairs."""

    def __init__(self, size=0, extensions=None):
        #: the chunk-size
        self.size = size
        #: declared extension parameters
        if extensions:
            self.extensions = self.bparameters(extensions)
        else:
            self.extensions = {}
        self._cx = sorted((p[0], p[1][1]) for p in dict_items(self.extensions))

    @classmethod
    def from_str(cls, source):
        """Parses the chunk header from a *source* string of *TEXT*.

        If the chunk header is not parsed correctly BadSyntax is raised.
        The header includes the chunk-size and any chunk-extension
        parameters but it does *not* include the trailing CRLF or the
        chunk-data"""
        p = ParameterParser(source)
        chunk = p.require_chunk()
        p.require_end("chunk")
        return chunk

    def to_bytes(self):
        return b"%X%s" % (self.size, format_parameters(self.extensions))

    def sortkey(self):
        return (self.size, self._cx)


class MediaType(SortableParameter):

    """Represents an HTTP media-type.

    The built-in str function can be used to format instances according
    to the grammar defined in the specification.

    type
        The type code string, defaults to 'application'

    subtype
        The sub-type code, defaults to 'octet-stream'

    parameters
        A dictionary such as would be returned by
        :py:meth:`grammar.WordParser.parse_parameters` containing the
        media type's parameters.

    Instances are immutable and support parameter value access by
    lower-case key (as a *character* string), returning the
    corresponding value or raising KeyError.  E.g., mtype['charset']

    Instances also define comparison methods and a hash implementation.
    Media-types are compared by (lower case) type, subtype and ultimately
    parameters."""

    def __init__(self, type="application", subtype="octet-stream",
                 parameters={}):
        self.type = type
        self.subtype = subtype
        if parameters:
            self.parameters = self.bparameters(parameters)
        else:
            self.parameters = {}
        self._hp = sorted((p[0], p[1][1]) for p in dict_items(self.parameters))

    @classmethod
    def from_str(cls, source):
        """Creates a media-type from a *source* string.

        Enforces the following rule from the specification:

            Linear white space (LWS) MUST NOT be used between the type
            and subtype, nor between an attribute and its value

        The source may be either characters or bytes.  Character strings
        must consist of iso-8859-1 characters only and should be plain
        ascii."""
        if is_unicode(source):
            source = source.encode('iso-8859-1')
        p = ParameterParser(source, ignore_sp=False)
        mt = p.require_media_type()
        p.parse_sp()
        p.require_end("media-type")
        return mt

    def to_bytes(self):
        return (''.join((self.type, '/', self.subtype)).encode('ascii') +
                format_parameters(self.parameters))

    def __repr__(self):
        return "MediaType(%s, %s, %s)" % (repr(self.type),
                                          repr(self.subtype),
                                          repr(self.parameters))

    def __getitem__(self, key):
        if key in self.parameters:
            return self.parameters[key][1]
        else:
            raise KeyError("MediaType instance has no parameter %s" %
                           repr(key))

    def __contains__(self, key):
        return key in self.parameters

    def sortkey(self):
        return (self.type.lower(), self.subtype.lower(), self._hp)


#: A predefined constant for application/octet-stream
APPLICATION_OCTETSTREAM = MediaType('application', 'octet-stream')

#: A predefined constant for text/plain
PLAIN_TEXT = MediaType('text', 'plain')


class ProductToken(SortableParameter):

    """Represents an HTTP product token.

    The comparison operations use a more interesting sort than plain
    text on version in order to provide a more intuitive ordering.  As
    it is common practice to use dotted decimal notation for versions
    (with some alphanumeric modifiers) the version string is exploded
    (see :py:meth:`explode`) internally on construction and this
    exploded value is used in comparisons.  The upshot is that version
    1.0.3 sorts before 1.0.10 as you would expect and 1.0a < 1.0 <
    1.0.3a3 < 1.0.3a20 < 1.0.3b1 < 1.0.3; there are limits to this
    algorithm.  1.0dev > 1.0b1 even though it looks like it should be
    the other way around.  Similarly 1.0-live < 1.0-prod etc.

    You shouldn't use this comparison as a definitive way to determine
    that one release is more recent or up-to-date than another unless
    you know that the product in question uses a numbering scheme
    compatible with these rules.  On the other hand, it can be useful
    when sorting lists for human consumption."""

    def __init__(self, token=None, version=None):
        #: the product's token
        self.token = token
        #: the product's version
        self.version = version
        if self.version is not None:
            # explode the version string
            self._version = self.explode(self.version)
        else:
            self._version = ()

    @classmethod
    def explode(cls, version):
        """Returns an exploded version string.

        Version strings are split by dot and then by runs of non-digit
        characters resulting in a list of tuples.  Numbers that have
        modified are treated as if they had a ~ suffix.  This ensures
        that when sorting, 1.0 > 1.0a (i.e., qualifiers indicate earlier
        releases, ~ being the ASCII character with the largest
        codepoint).

        Examples will help::

                explode("2.15")==((2, "~"),(15, "~"))
                explode("2.17b3")==((2, "~"),(17, "b", 3, "~"))
                explode("2.b3")==((2, "~"),(-1, "b", 3, "~"))

        Note that a missing leading numeric component is treated as -1
        to force "a3" to sort before "0a3"."""
        exploded = []
        p = BasicParser(force_text(version))
        while p.the_char is not None:
            # parse an item
            vitem = []
            modifier = []
            while not p.match(".") and not p.match_end():
                num = p.parse_integer()
                if num is None:
                    if not vitem:
                        vitem.append(-1)
                    modifier.append(p.the_char)
                    p.next_char()
                else:
                    if modifier:
                        vitem.append("".join(modifier))
                        modifier = []
                    vitem.append(num)
            if modifier:
                vitem.append("".join(modifier))
            elif vitem:
                # forces 1.0 > 1.0b
                vitem.append('~')
            exploded.append(tuple(vitem))
            p.parse(".")
        return tuple(exploded)

    @classmethod
    def from_str(cls, source):
        """Creates a product token from a *source* string."""
        p = ParameterParser(source)
        p.parse_sp()
        pt = p.require_product_token()
        p.parse_sp()
        p.require_end("product token")
        return pt

    @classmethod
    def list_from_str(cls, source):
        """Creates a list of product tokens from a *source* string.

        Individual tokens are separated by white space."""
        ptlist = []
        p = ParameterParser(source)
        p.parse_sp()
        while p.the_word is not None:
            ptlist.append(p.require_product_token())
            p.parse_sp()
        p.require_end("product token")
        return ptlist

    def to_bytes(self):
        if self.version is None:
            return self.token.encode('ascii')
        else:
            return ''.join((self.token, '/', self.version)).encode('ascii')

    def __repr__(self):
        return "ProductToken(%s, %s)" % (repr(self.token), repr(self.version))

    def sortkey(self):
        return (self.token, self._version)


class LanguageTag(SortableParameter):

    """Represents an HTTP language-tag.

    Language tags are compared by lower casing all components and then
    sorting by primary tag, then by each sub-tag. Note that en sorts
    before en-US."""

    def __init__(self, primary, *subtags):
        self.primary = primary
        self.subtags = tuple(subtags)
        tags = [primary.lower()]
        for sub in subtags:
            tags.append(sub.lower())
        self._tag = tuple(tags)

    def partial_match(self, range):
        """True if this tag is a partial match against *range*

        range
            A tuple of lower-cased subtags.  An empty tuple matches all
            instances.

        For example::

            lang=LanguageTag("en",("US","Texas"))
            lang.partial_match(())==True
            lang.partial_match(("en",)==True
            lang.partial_match(("en","us")==True
            lang.partial_match(("en","us","texas")==True
            lang.partial_match(("en","gb")==False
            lang.partial_match(("en","us","tex")==False"""
        if len(range) > len(self._tag):
            return False
        for i in range3(len(range)):
            if self._tag[i] != range[i]:
                return False
        return True

    @classmethod
    def from_str(cls, source):
        """Creates a language tag from a *source* string.

        Enforces the following rules from the specification:

                White space is not allowed within the tag"""
        p = ParameterParser(source, ignore_sp=False)
        p.parse_sp()
        t = p.require_language_tag()
        p.parse_sp()
        p.require_end("language tag")
        return t

    @classmethod
    def list_from_str(cls, source):
        """Creates a list of language tags from a *source* string."""
        p = ParameterParser(source, ignore_sp=False)
        tags = []
        while True:
            p.parse_sp()
            t = p.parse_production(p.require_language_tag)
            if t is not None:
                tags.append(t)
            p.parse_sp()
            if not p.parse_separator(","):
                break
        p.parse_sp()
        p.require_end("language tag")
        return tags

    def to_bytes(self):
        return '-'.join([self.primary] + list(self.subtags)).encode('ascii')

    def __repr__(self):
        return (
            "LanguageTag(%s)" % ','.join(
                [repr(tag) for tag in [self.primary] + list(self.subtags)]))

    def sortkey(self):
        return self._tag


class EntityTag(SortableParameter):

    """Represents an HTTP entity-tag.

    tag
        The opaque tag

    weak
        A boolean indicating if the entity-tag is a weak or strong
        entity tag.  Defaults to True.

    Instances are compared by tag and then, if the tags match, by
    wheather the tag is weak or not."""

    def __init__(self, tag, weak=True):
        #: True if this is a weak tag
        self.weak = weak
        #: the opaque tag
        self.tag = self.bstr(tag)

    @classmethod
    def from_str(cls, source):
        """Creates an entity-tag from a *source* string."""
        p = ParameterParser(source)
        et = p.require_entity_tag()
        p.require_end("entity-tag")
        return et

    def to_bytes(self):
        if self.weak:
            return b"W/" + quote_string(self.tag)
        else:
            return quote_string(self.tag)

    def __repr__(self):
        return ("EntityTag(%s,%s)" %
                (repr(self.tag), "True" if self.week else "False"))

    def sortkey(self):
        return (self.tag, self.weak)


SLASH = byte('/')
COLON = byte(':')


class ParameterParser(WordParser):

    """An extended parser for parameter values

    This parser defines attributes for dealing with English date names
    that are useful beyond the basic parsing functions to allow the
    formatting of date information in English regardless of the
    locale."""

    def require_http_version(self):
        """Parses an :py:class:`HTTPVersion` instance

        Returns an :class:`HTTPVersion` instance."""
        self.parse_sp()
        token = self.require_token(b"HTTP").upper()
        if token != b"HTTP":
            self.parser_error('HTTP')
        self.parse_sp()
        self.require_separator(SLASH, "HTTP/")
        self.parse_sp()
        token = self.require_token("protocol version")
        version = token.split(b'.')
        if (len(version) != 2 or not is_digits(version[0]) or
                not is_digits(version[1])):
            self.parser_error('version')
        return HTTPVersion(major=int(version[0]), minor=int(version[1]))

    #: A list of English day-of-week abbreviations: wkday[0] == "Mon",
    #: etc.
    wkday = [b"Mon", b"Tue", b"Wed", b"Thu", b"Fri", b"Sat", b"Sun"]

    #: A list of English day-of-week full names: weekday[0] == "Monday"
    weekday = [b"Monday", b"Tuesday", b"Wednesday", b"Thursday", b"Friday",
               b"Saturday", b"Sunday"]

    _wkdayTable = {}

    #: A list of English month names: month[0] == "Jan", etc.
    month = [b"Jan", b"Feb", b"Mar", b"Apr", b"May", b"Jun",
             b"Jul", b"Aug", b"Sep", b"Oct", b"Nov", b"Dec"]

    _monthTable = {}

    def require_fulldate(self):
        """Parses a :py:class:`FullDate` instance.

        Returns a :class:`FullDate` instance or raises
        :py:class:`~pyslet.http.grammar.BadSyntax` if none is found."""
        century = None
        year = None
        month = None
        day = None
        hour = None
        minute = None
        second = None
        dayofweek = None
        self.parse_sp()
        token = self.require_token("day-of-week").lower()
        dayofweek = self._wkdayTable.get(token, None)
        if dayofweek is None:
            self.parser_error("day of week")
        self.parse_sp()
        if self.parse_separator(COMMA):
            self.parse_sp()
            if self.is_token() and is_digits(self.the_word):
                # Best format 0: "Sun, 06 Nov 1994 08:49:37 GMT" - the
                # preferred format!
                day = self.require_integer("date")
                self.parse_sp()
                token = self.require_token("month").lower()
                month = self._monthTable.get(token, None)
                if month is None:
                    self.parser_error("month")
                self.parse_sp()
                year = self.require_integer("year")
                century = year // 100
                year = year % 100
            else:
                # Alternative 1: "Sunday, 06-Nov-94 08:49:37 GMT"
                token = self.require_token("DD-MMM-YY")
                stoken = token.split(b'-')
                if (len(stoken) != 3 or not is_digits(stoken[0]) or
                        not is_digits(stoken[2])):
                    self.parser_error("DD-MMM-YY")
                day = int(stoken[0])
                year = int(stoken[2])
                month = self._monthTable.get(stoken[1].lower(), None)
                if month is None:
                    self.parser_error("month")
        else:
            # "Sun Nov  6 08:49:37 1994"
            token = self.require_token("month").lower()
            month = self._monthTable.get(token, None)
            if month is None:
                self.parser_error("month")
            self.parse_sp()
            day = self.require_integer("date")
        self.parse_sp()
        hour = self.require_integer("hour")
        self.require_separator(COLON)
        minute = self.require_integer("minute")
        self.require_separator(COLON)
        second = self.require_integer("second")
        self.parse_sp()
        if year is None:
            year = self.require_integer("year")
            century = year // 100
            year = year % 100
        else:
            token = self.require_token("GMT").upper()
            if token != b"UTC" and token != b"GMT":
                self.parser_error("timezone")
        if century is None:
            if year < 90:
                century = 20
            else:
                century = 19
        tp = FullDate(
            date=iso.Date(century=century, year=year, month=month + 1,
                          day=day),
            time=iso.Time(hour=hour, minute=minute, second=second,
                          zdirection=0))
        d1, d2, d3, d4, dow = tp.date.get_week_day()
        if dow != dayofweek + 1:
            self.parser_error("matching day-of-week")
        return tp

    #: Parses a delta-seconds value, see
    #: :py:meth:`WordParser.parse_integer`
    parse_delta_seconds = WordParser.parse_integer

    #: Parses a charset, see :py:meth:`WordParser.parse_tokenlower`
    parse_charset = WordParser.parse_tokenlower

    #: Parses a content-coding, see
    #: :py:meth:`WordParser.parse_tokenlower`
    parse_content_coding = WordParser.parse_tokenlower

    def require_transfer_encoding(self):
        """Parses a :py:class:`TransferEncoding` instance"""
        self.parse_sp()
        token = self.require_token(
            "transfer-encoding").lower().decode('iso-8859-1')
        if token != "chunked":
            parameters = {}
            self.parse_parameters(parameters)
            return TransferEncoding(token, parameters)
        else:
            return TransferEncoding()

    def require_chunk(self):
        """Parses a chunk header

        Returns a :py:class:`Chunk` instance."""
        self.parse_sp()
        size = self.require_hexinteger("chunk-size")
        extensions = {}
        self.parse_parameters(extensions)
        return Chunk(size, extensions)

    def require_media_type(self):
        """Parses a :py:class:`MediaType` instance."""
        self.parse_sp()
        type = self.require_token("media-type").lower().decode('ascii')
        self.require_separator(SLASH, "media-type")
        subtype = self.require_token("media-subtype").lower().decode('ascii')
        self.parse_sp()
        parameters = {}
        self.parse_parameters(parameters, ignore_allsp=False)
        return MediaType(type, subtype, parameters)

    def require_product_token(self):
        """Parses a :py:class:`ProductToken` instance.

        Raises BadSyntax if no product token was found."""
        self.parse_sp()
        token = self.require_token("product token").decode('ascii')
        self.parse_sp()
        if self.parse_separator(SLASH):
            version = self.require_token("product-version").decode('ascii')
        else:
            version = None
        return ProductToken(token, version)

    def parse_qvalue(self):
        """Parses a qvalue returning a float

        Returns None if no qvalue was found."""
        if self.is_token():
            q = None
            qsplit = self.the_word.split(b'.')
            if len(qsplit) == 1:
                if is_digits(qsplit[0]):
                    q = float(qsplit[0])
            elif len(qsplit) == 2:
                if is_digits(qsplit[0]) and is_digits(qsplit[1]):
                    q = float("%.3f" % float(self.the_word))
            if q is None:
                return None
            else:
                if q > 1.0:
                    # be generous if the value has overflowed
                    q = 1.0
                self.parse_word()
                return q
        else:
            return None

    def require_language_tag(self):
        """Parses a language tag returning a :py:class:`LanguageTag`
        instance.  Raises BadSyntax if no language tag was
        found."""
        self.parse_sp()
        tag = [t.decode('ascii') for t in
               self.require_token("language-tag").split(b'-')]
        self.parse_sp()
        return LanguageTag(tag[0], *tag[1:])

    def require_entity_tag(self):
        """Parses an entity-tag returning a :py:class:`EntityTag`
        instance.  Raises BadSyntax if no language tag was
        found."""
        self.parse_sp()
        w = self.parse_token()
        self.parse_sp()
        if w is not None:
            if w.upper() != b"W":
                self.parser_error("entity-tag")
            self.require_separator(SLASH, "entity-tag")
            self.parse_sp()
            w = True
        else:
            w = False
        tag = self.require_production(self.parse_quoted_string(), "entity-tag")
        self.parse_sp()
        return EntityTag(tag, w)


for i in range3(len(ParameterParser.wkday)):
    ParameterParser._wkdayTable[ParameterParser.wkday[i].lower()] = i
for i in range3(len(ParameterParser.weekday)):
    ParameterParser._wkdayTable[ParameterParser.weekday[i].lower()] = i
for i in range3(len(ParameterParser.month)):
    ParameterParser._monthTable[ParameterParser.month[i].lower()] = i



#! /usr/bin/env python

import warnings

from .xml.xsdatatypes import *  	# noqa

warnings.warn(
    "pyslet.xsdatatypes20041029 is deprecated, use pyslet.xml.xsdatatypes "
    "instead", DeprecationWarning, stacklevel=3)



print("import: 'pyslet'")
import pyslet

print("import: 'pyslet.http'")
import pyslet.http

print("import: 'pyslet.odata2'")
import pyslet.odata2

print("import: 'pyslet.qtiv1'")
import pyslet.qtiv1

print("import: 'pyslet.qtiv2'")
import pyslet.qtiv2

print("import: 'pyslet.xml20081126'")
import pyslet.xml20081126




#! /usr/bin/env python

import time
import logging
import threading
import os.path

from .. import iso8601 as iso
from ..py2 import (
    byte,
    byte_value,
    dict_values,
    is_unicode,
    join_bytes,
    to_bytes)

from . import grammar
from . import params


def is_cookie_octet(b):
    """Tests a byte against production coookie_octet"""
    if b is not None:
        return 0x20 < byte_value(b) < 0x7F and b not in b'",;\\'
    else:
        return False

HYPHEN = byte('-')
FULL_STOP = byte('.')


def is_ldh_label(label):
    """Tests a binary string against the definition of LDH label

    LDH Label is defined in RFC5890_ as being the classic label syntax
    defined in RFC1034_ and updated in RFC1123_.  To cut a long story
    short the update in question is described as follows:

        One aspect of host name syntax is hereby changed: the
        restriction on the first character is relaxed to allow either a
        letter or a digit.

    ..  _RFC5890: http://tools.ietf.org/html/rfc5890

    ..  _RFC1034: http://tools.ietf.org/html/rfc1034

    ..  _RFC1123: http://tools.ietf.org/html/rfc1123

    Although not spelled out there this would make the updated syntax::

        <label> ::= <let-dig> [ [ <ldh-str> ] <let-dig> ]
        <ldh-str> ::= <let-dig-hyp> | <let-dig-hyp> <ldh-str>
        <let-dig-hyp> ::= <let-dig> | "-"
        <let-dig> ::= <letter> | <digit>"""
    allow_hyphen = False
    trailing_hyphen = False
    if not label or len(label) > 63:
        return False
    for b in label:
        if (grammar.is_alpha(b) or grammar.is_digit(b) or
                (allow_hyphen and b == HYPHEN)):
            trailing_hyphen = (b == HYPHEN)
            allow_hyphen = True
        else:
            return False
    return not trailing_hyphen


def is_rldh_label(label):
    """Tests a binary string against the definition of R-LDH label

    As defined by RFC5890_

        Reserved LDH labels, known as "tagged domain names" in some
        other contexts, have the property that they contain "--" in the
        third and fourth characters but which otherwise conform to LDH
        label rules.

        Non-Reserved LDH labels are the set of valid LDH labels that do
        not have "--" in the third and fourth positions.

    Therefore you can test for a NR-LDH label simply by using the *not*
    operator."""
    return is_ldh_label(label) and label[2:4] == b'--'


def is_a_label(label):
    """Test a binary string against the definition of A-label.

    As defined by RFC5890_

    In fact, this function currently only tests for being an XN--
    label.

        the class of labels that begin with the prefix "xn--" (case
        independent), but otherwise conform to the rules for LDH labels
        [is called "XN-labels"]...

        The XN-labels that are valid Punycode output are known as
        "A-labels" if they also meet the other criteria for IDNA-validity

    So bear in mind that (a) the remainder of the label may fail to
    decode properly when passed to the punycode algorithm and (b) even
    if it does decode it may result in a string that is not actually a
    valid U-Label."""
    return is_ldh_label(label) and label[:4].lower() == b'xn--'


def split_domain(domain_str, allow_wildcard=False):
    """Splits a domain string

    domain_str
        A character string, or a UTF-8 encoded binary string.

    allow_wildcard (Default: False)
        Allows the use of a single '*' character as a domain label for
        the purposes of parsing wildcard domain definitions.

    Returns a list of lower cased *ASCII* labels as character strings,
    converting U-Labels to ACE form (xn--) in the process.  For example::

        >>> split_domain('example.COM')
        >>> ['example', 'com']
        >>> split_domain(u'\\u98df\\u72ee.com.cn')
        >>> ['xn--85x722f', 'com', 'cn']

    Raises ValueError if domain_str is not valid."""
    if is_unicode(domain_str):
        domain_str = domain_str.encode('utf-8')
    domain = domain_str.split(b'.')
    result = []
    for label in domain:
        if is_ldh_label(label):
            result.append(label.lower())
        elif allow_wildcard and label == b'*':
            result.append(b'*')
        else:
            # some high-bit values
            try:
                alabel = b"xn--%s" % label.decode('utf-8').encode('punycode')
                if not is_ldh_label(alabel):
                    raise ValueError("bad label in domain: %s" % domain_str)
            except UnicodeError:
                raise ValueError("bad label in domain: %s" % repr(domain_str))
            result.append(alabel.lower())
    return [r.decode('ascii') for r in result]


def domain_in_domain(subdomain, domain):
    """Returns try if subdomain is a sub-domain of domain.

    subdomain
        A *reversed* list of strings returned by :func:`split_domain`

    domain
        A *reversed* list of strings as returned by :func:`split_domain`

    For example::

        >>> domain_in_domain(['com', 'example'],
        ...                  ['com', 'example', 'www'])
        True"""
    if len(subdomain) <= len(domain):
        i = 0
        for d in subdomain:
            if d != domain[i]:
                return False
            i += 1
        return True
    else:
        return False


def encode_domain(domain, allow_wildcard=False):
    """Returns domain correctly encoded as a binary string

    domain
        A binary or character string containing a representation of a
        domain using either U-Labels or ACE form for non-ASCII
        characters.

    allow_wildcard (Default: False)
        Allows the use of a single '*' character as a domain label for
        the purposes of encoding wildcard domain definitions.

    The result is a character string containing only ASCII encoded
    characters."""
    return '.'.join(split_domain(domain, allow_wildcard))


def is_delimiter(b):
    """Tests a character against the production delimiter

    This production is from the weaker section 5 syntax of RFC6265."""
    if b is not None:
        b = byte_value(b)
        return (b == 0x09 or
                (0x20 <= b <= 0x2F) or (0x3B <= b <= 0x40) or
                (0x5B <= b <= 0x60) or (0x7B <= b <= 0x7E))
    else:
        return False


def is_non_delimiter(b):
    """Tests a character against the production non-delimiter

    The result differs from using *not is_delimiter* only in the
    handling of None which will return False when passed to either
    function."""
    if b is not None:
        return not is_delimiter(b)
    else:
        return False


def is_non_digit(b):
    """Tests a character against the production non-digit."""
    if b is not None:
        return not grammar.is_digit(b)
    else:
        return False


def split_year(year_str):
    """Parses a year from a binary string

    Uses the generous rules in section 5.1 and returns a year value,
    adjusted using the 2-digit year algorithm documented there.

    If a year value can't be found ValueError is raised."""
    digits = []
    for b in year_str:
        if grammar.is_digit(b):
            digits.append(b)
            if len(digits) > 4:
                raise ValueError("Can't read year from %s" % year_str)
        else:
            break
    if len(digits) < 2:
        raise ValueError("Can't read year from %s" % year_str)
    year = int(join_bytes(digits))
    if year > 69 and year < 100:
        year += 1900
    elif year < 70:
        year += 2000
    return year


MONTHS = {
    b"jan": 1,
    b"feb": 2,
    b"mar": 3,
    b"apr": 4,
    b"may": 5,
    b"jun": 6,
    b"jul": 7,
    b"aug": 8,
    b"sep": 9,
    b"oct": 10,
    b"nov": 11,
    b"dec": 12}


def split_month(month_str):
    """Parses a month from a string

    Uses the generous rules in section 5.1 and returns a month value
    from 1 (January) to 12 (December).

    If a month value can't be found ValueError is raised."""
    try:
        return MONTHS[month_str[0:3].lower()]
    except KeyError:
        raise ValueError("Can't read month from %s" % month_str)


def split_day_of_month(dom_str):
    """Parses a day-of-month from a binary string

    Users the generous rules in section 5.1 and returns a single integer
    or raises ValueError if a valid day of month can't be found."""
    digits = []
    for b in dom_str:
        if grammar.is_digit(b):
            digits.append(b)
            if len(digits) > 2:
                raise ValueError("Can't read day-of-month from %s" % dom_str)
        else:
            break
    if len(digits) < 1:
        raise ValueError("Can't read day-of-month from %s" % dom_str)
    return int(join_bytes(digits))


def split_time(time_str):
    """Parses a time from a binary string

    Users the generous rules in section 5.1 and returns a triple
    of hours, minutes, seconds.  These values are unchecked!

    If the time can't be found ValueError is raised."""
    components = []
    digits = []
    for b in time_str:
        if grammar.is_digit(b):
            digits.append(b)
            if len(digits) > 2:
                raise ValueError("Can't read time from %s" % time_str)
        else:
            if len(digits) < 1:
                raise ValueError("Missing time-value in %s" % time_str)
            components.append(int(join_bytes(digits)))
            digits = []
            if len(components) == 3 or b != grammar.COLON:
                break
    if digits:
        components.append(int(join_bytes(digits)))
    if (len(components) != 3):
        raise ValueError("Can't read time from %s" % time_str)
    return tuple(components)


class CookieError(ValueError):

    """Raised when an operation violates RFC6265_ rules."""
    pass


class CookieParser(grammar.OctetParser):

    """General purpose class for parsing RFC6265_ productions

    Unlike the basic syntax functions these methods allow a longer
    string, such as that received from an HTTP header, to be parsed into
    its component parts.

    Methods follow inherited naming conventions, require\\_ methods
    raise a ValueError if the production is not matched whereas parse\\_
    methods optionally parse a production if it is present and return
    None if not present."""

    def require_set_cookie_string(self, strict=False):
        """Parses the set-cookie-string production

        strict (Default: False)
            Use the stricter section 4 syntax rules instead of the more
            permissive algorithm described in section 5.2

        This is the format of the Set-Cookie header, it returns a
        :class:`Cookie` instance or None if this cookie definition
        should be ignored."""
        if strict:
            return self._require_set_cookie_string_strict()
        try:
            name, value = self.require_name_value_pair()
        except ValueError:
            name, value = None, None
        if self.parse(b';'):
            attrs = self.parse_until(None)
        else:
            attrs = None
        if not name:
            # ignore this definition!
            return None
        expires = max_age = domain = path = None
        secure = http_only = False
        while attrs is not None:
            pos = attrs.find(b';')
            if pos < 0:
                av = attrs
                attrs = None
            else:
                av = attrs[:pos]
                attrs = attrs[pos + 1:]
            pos = av.find(b'=')
            if pos < 0:
                aname = av
                avalue = ''
            else:
                aname = av[:pos].strip()
                avalue = av[pos + 1:].strip()
            aname_lower = aname.lower()
            if aname_lower == b"expires":
                p = CookieParser(avalue)
                expires = p.require_cookie_date()
            elif aname_lower == b"max-age":
                if avalue and avalue[0] == HYPHEN:
                    if len(avalue) == 1:
                        avalue = b"0"
                    else:
                        for c in avalue[1:]:
                            if not grammar.is_digit(c):
                                avalue = ''
                                break
                else:
                    for c in avalue:
                        if not grammar.is_digit(c):
                            avalue = ''
                            break
                if avalue:
                    max_age = int(avalue)
            elif aname_lower == b'domain':
                if avalue:
                    if avalue[0] == FULL_STOP:
                        domain = avalue[1:].lower()
                    else:
                        domain = avalue.lower()
                    domain = Cookie.cstr(domain)
            elif aname_lower == b"path":
                if avalue and avalue[0] == grammar.SOLIDUS:
                    path = Cookie.cstr(avalue)
            elif aname_lower == b"secure":
                secure = True
            elif aname_lower == b"httponly":
                http_only = True
        return Cookie(name, value, expires=expires, max_age=max_age,
                      domain=domain, path=path, secure=secure,
                      http_only=http_only)

    def require_cookie_string(self, strict=False):
        """Parses the value of a Cookie header.

        strict (Default: False)
            Indicates if stricter section 4 parsing is required.

        Returns a dictionary of values, the keys are the names of the
        cookies in the cookie string and the values are either strings
        or, in the case of multiply defined names, sets of strings.  We
        use sets as the specification makes it clear that you should not
        rely on the order of such definitions.  All strings (including
        cookie names) are binary strings."""
        while True:
            # loop around in case of double folds, this is the
            # difference between OWS and LWS
            if not self.parse_lws():
                break
        cookie_list = {}
        try:
            savepos = self.pos
            while True:
                if strict:
                    name, value = self.require_cookie_pair()
                else:
                    name, value = self.require_name_value_pair()
                if name in cookie_list:
                    values = cookie_list[name]
                    if not isinstance(values, set):
                        values = set((values, ))
                        cookie_list[name] = values
                    values.add(value)
                else:
                    cookie_list[name] = value
                savepos = self.pos
                if self.parse(b';'):
                    # bit generous here, accept multiple spaces
                    if strict:
                        self.require(b' ')
                    else:
                        while self.the_char is not None:
                            if not self.parse_one(b' \t'):
                                break
                    continue
                else:
                    break
            while True:
                # more OWS
                if not self.parse_lws():
                    break
        except ValueError:
            self.setpos(savepos)
        return cookie_list

    def _require_set_cookie_string_strict(self):
        name, value = self.require_cookie_pair()
        expires = max_age = domain = path = None
        secure = http_only = False
        extensions = []
        while self.parse(b';'):
            # we just parsed a ';' which must be followed by SP
            if not self.parse(b' '):
                self.setpos(self.pos - 1)
                break
            if self.parse_insensitive(b'expires='):
                sane_date = self.parse_sane_cookie_date()
                if ((self.the_char is None or
                        self.the_char == grammar.SEMICOLON) and
                        sane_date is not None):
                    expires = sane_date
                    continue
            elif self.parse_insensitive(b'max-age='):
                digits = self.parse_digits(min=1)
                if digits and digits[0] != '0':
                    max_age = int(digits)
                    continue
            elif self.parse_insensitive(b'domain='):
                domain = self.parse_until(b';')
                domain = Cookie.cstr(domain)
                continue
            elif self.parse_insensitive(b'path='):
                path = Cookie.cstr(self.parse_until(b';'))
                continue
            elif self.parse_insensitive(b'secure'):
                if self.the_char is None or self.the_char == grammar.SEMICOLON:
                    secure = True
                    continue
            elif self.parse_insensitive(b'httponly'):
                if self.the_char is None or self.the_char == grammar.SEMICOLON:
                    http_only = True
                    continue
            # extension
            extensions.append(self.parse_cookie_av())
        return Section4Cookie(
            name, value, expires=expires, max_age=max_age,
            domain=domain, path=path, secure=secure,
            http_only=http_only, extensions=extensions)

    def require_name_value_pair(self):
        """Returns a (name, value) pair

        Parsed according to the looser section 5 syntax so will allow
        almost anything as a name and value provided it has an '='."""
        nvpair = self.parse_until(b';')
        pos = nvpair.find(b'=')
        if pos < 0:
            raise ValueError("Expected '=' in name-value-pair: %s" % nvpair)
        name = nvpair[:pos].strip()
        value = nvpair[pos + 1:].strip()
        return name, value

    def require_cookie_pair(self):
        """Returns a (name, value) pair parsed according to cookie-pair

        Parsed according to the stricter section 4 syntax so will only
        accept valid tokens as names, the '=' is required and the value
        must be parseable with :meth:`require_cookie_value`."""
        name = self.require_production(self.parse_token())
        self.require(b'=')
        value = self.require_cookie_value()
        return name, value

    def parse_cookie_pair(self):
        """See: :meth:`require_cookie_pair`

        If not parsed returns (None, None) rather than just None."""
        result = self.parse_production(self.require_cookie_pair)
        if result is None:
            return None, None
        else:
            return result

    def require_cookie_value(self):
        """Returns a cookie-value (binary) string.

        Parsed according to the stricter section 4 syntax so will not
        allow whitespace, comma, semicolon or backslash characters and
        will only allow double-quote when it is used to completely
        "enclose" the value, in which case the double-quotes are still
        considered to be part of the value string."""
        value = []
        if self.parse(b'"'):
            dqflag = True
        else:
            dqflag = False
        while self.the_char is not None and is_cookie_octet(self.the_char):
            value.append(self.the_char)
            self.next_char()
        if dqflag:
            self.require(b'"')
            value[0:0] = (grammar.DQUOTE, )
            value.append(grammar.DQUOTE)
            value = join_bytes(value)
        else:
            value = join_bytes(value)
        return value

    def parse_cookie_av(self):
        """Parses a cookie-av string.

        This production is effectively the production for extension-av
        in the stricter section 4 syntax.  Effectively it returns
        everything up to but not including the next ';' or CTL
        character.

        It never returns None, if nothing is found it returns an empty
        string instead."""
        av = []
        while self.the_char is not None and self.the_char != grammar.SEMICOLON:
            if grammar.is_ctl(self.the_char):
                break
            else:
                av.append(self.the_char)
                self.next_char()
        return join_bytes(av)

    def parse_sane_cookie_date(self):
        """Parses the sane-cookie-date production.

        This is the stricter syntax defined in section 4.  The returns
        result is a :class:`~params.FullDate` instance."""
        savepos = self.pos
        date_str = self.parse_until(b'GMT')
        date_str = date_str + self.parse(b'GMT')
        try:
            return params.FullDate.from_http_str(date_str)
        except ValueError:
            self.setpos(savepos)
            return None

    def parse_cookie_date_tokens(self):
        """Parses a date-token-list

        This uses the weak section 5.1 syntax

        It never returns None, if there are no tokens then it returns an
        empty list.  Delimiters are always discarded."""
        token_list = []
        while is_delimiter(self.the_char):
            self.next_char()
        while self.the_char is not None:
            token = []
            while is_non_delimiter(self.the_char):
                token.append(self.the_char)
                self.next_char()
            token_list.append(join_bytes(token))
            while is_delimiter(self.the_char):
                self.next_char()
        return token_list

    def require_cookie_date(self):
        """Parses a date value

        This uses the weak section 5.1 syntax and the algorithm
        described there.  It absorbs almost all errors returning None if
        this date value should be ignored - but warnings are logged to
        alert you to the failure.  The implications of replacing a date
        with None in this syntax are typically that a cookie that is
        supposed to be persistent become session only.  However, if this
        was an attempt to remove a cookie with a very early date then
        the failure could cause more problems.

        If successful, it returns a :class:`~params.FullDate`
        instance."""
        token_list = self.parse_cookie_date_tokens()
        time_str = None
        dom_str = None
        month_str = None
        year_str = None
        hour = minute = second = day = month = year = None
        for token in token_list:
            if not time_str:
                try:
                    hour, minute, second = split_time(token)
                    time_str = token
                    continue
                except ValueError:
                    pass
            if not dom_str:
                try:
                    day = split_day_of_month(token)
                    dom_str = token
                    continue
                except ValueError:
                    pass
            if not month_str:
                try:
                    month = split_month(token)
                    month_str = token
                    continue
                except ValueError:
                    pass
            if not year_str:
                try:
                    year = split_year(token)
                    year_str = token
                    continue
                except ValueError:
                    pass
        if not time_str:
            logging.warning("No time found in %s", b''.join(token_list))
            return None
        if not dom_str:
            logging.warning(
                "No day of month found in %s", b''.join(token_list))
            return None
        if not month_str:
            logging.warning("No month found in %s", b''.join(token_list))
            return None
        if not year_str:
            logging.warning("No year found in %s", b''.join(token_list))
            return None
        if day < 1 or day > 31:
            logging.warning("Bad day found in %s", b''.join(token_list))
            return None
        if year < 1601:
            logging.warning("Cookie too old: %s", b''.join(token_list))
            return None
        if hour > 23 or minute > 59 or second > 59:
            logging.warning("Bad time found in: %s", b''.join(token_list))
            return None
        century = year // 100
        year = year % 100
        try:
            return params.FullDate(
                date=iso.Date(century=century, year=year, month=month,
                              day=day),
                time=iso.Time(hour=hour, minute=minute, second=second,
                              zdirection=0))
        except iso.DateTimeError:
            logging.warning("Bad cookie date: %s", b''.join(token_list))
            return None


class Cookie(params.Parameter):

    """Represents the definition of a cookie

    Where binary strings are required, character strings will be
    accepted and converted using UTF-8 but non-ASCII characters are
    *not* portable and should be avoided.

    name
        The name of the cookie as a binary string.

    value
        The value of the cookie as a binary string.

    path (optional)
        A character string containing the path of the cookie.  If None
        then the 'directory' of the page that returned the cookie will
        be used by the client.

    domain (optional)
        A character string containing the domain of the cookie.  If None
        then the host name of the server that returned the cookie will
        be used by the client and the cookie will be treated as 'host
        only'.

    expires (optional)
        An :class:`~pyslet.iso8601.TimePoint` instance.  If None then the
        cookie will be treated as a session cookie by the client.

    max_age (optional)
        An integer, the length of time before the cookie expires in
        seconds.  Overrides the expires value.  If None then the value
        of expires is used instead, if both are None then the cookie
        will be treated as a session cookie by the client.

    secure (Default: False)
        Whether or not the cookie should be exposed only over secure
        protocols, such as https.

    http_only (Default: False)
        Whether or not the cookie should be exposed only via the HTTP
        protocol.  Recommended value: True!

    extensions
        A list of binary strings containing attribute extensions.  The
        strings should be of the form name=value but this is not
        enforced.

    Instances can be converted to strings using the builtin str function
    and the output that results is a valid Set-Cookie header value."""

    def __init__(self, name, value, path=None, domain=None, expires=None,
                 max_age=None, secure=False, http_only=False, extensions=None):
        now = time.time()
        #: the cookie's name
        self.name = self.bstr(name)
        #: the cookie's name
        self.value = self.bstr(value)
        #: the cookie's path
        self.path = path
        if domain:
            if domain[0:1] == '.':
                # strip leading '.'
                domain = domain[1:]
            elif domain[-1:] == '.':
                domain = None
        #: the cookie's domain
        self.domain = domain
        #: the cookie's secure flag
        self.secure = secure
        #: the cookie's httponly flag
        self.http_only = http_only
        #: the creation time of the cookie, initialised to the current
        #: time as returned by the builtin time.time function.
        self.creation_time = now
        #: the last access time of the cookie, initialised to the
        #: current time as returned by the builtin time.time function.
        self.access_time = now
        if max_age is None:
            #: the max_age value
            self.max_age = None
            if expires is None:
                #: the expiry time of the cookie, as an integer
                #: compatible with the value returned by time.time
                self.expires_time = None
            else:
                self.expires_time = expires.get_unixtime()
        else:
            if max_age <= 0:
                self.expires_time = 0
            else:
                self.expires_time = now + max_age
            self.max_age = max_age
        #: the expires value as passed to the constructor, this is
        #: preserved and is used when serialising the definition even if
        #: Max-Age is also in effect.  Some older clients may not
        #: support Max-Age and they will look at the Expires time
        #: instead.
        self.expires = expires
        #: the list of extensions
        if extensions:
            self.extensions = [self.bstr(e) for e in extensions]
        else:
            self.extensions = None

    @classmethod
    def bstr(cls, arg):
        """Overridden to use UTF-8 for binary encoding of arguments

        This method is used to convert arguments provided in
        constructors to the binary strings required by some attributes.
        The default implementation uses ISO-8859-1 in keeping with the
        general HTTP specification (although no encoding is portable
        across a wide range of browser and server implementations).

        We override it here for Cookies because the cookie specification
        hints that UTF-8 would be an appropriate choice for displaying
        binary information found in cookies."""
        if is_unicode(arg):
            return arg.encode('utf-8')
        else:
            return arg

    @classmethod
    def cstr(cls, value):
        """Used to interpret binary strings in cookies

        This method can be used as a default way of interpreting binary
        information found in cookies.  It tries to decode using UTF-8
        but, if that fails, it reverts to the default HTTP encoding of
        ISO-8859-1.  It is used sparingly, in most cases binary values
        are left uninterpreted but attributes such as the path and
        domain must be interpreted in relation to components of the URL
        and URLs use characters.

        For clarity, a domain name containing non-ASCII characters
        (U-labels) that has simply been UTF-8 encoded will be converted
        back to the original form (with U-labels) whereas the same
        domain correctly encoded in ACE format (xn--) will be unchanged
        by this decoding."""
        try:
            return value.decode('utf-8')
        except UnicodeDecodeError:
            return value.decode('iso-8859-1')

    def to_bytes(self):
        components = [b"%s=%s" % (self.name, self.value)]
        if self.path:
            components.append(b"Path=%s" % self.path.encode('ascii'))
        if self.domain:
            components.append(b"Domain=%s" %
                              encode_domain(self.domain).encode('ascii'))
        if self.expires is not None:
            components.append(b"Expires=%s" % to_bytes(self.expires))
        if self.max_age is not None:
            components.append(b"Max-Age=%s" % to_bytes(self.max_age))
        if self.secure:
            components.append(b"Secure")
        if self.http_only:
            components.append(b"HttpOnly")
        if self.extensions:
            components = components + self.extensions
        return b'; '.join(components)

    @classmethod
    def from_str(cls, src):
        """Creates a new instance from a src string

        The string is parsed using the generous parsing rules of Section
        5 of the specification.  Returns a new instance."""
        p = CookieParser(src)
        result = p.require_set_cookie_string()
        p.require_end("Set-Cookie string")
        return result

    def is_persistent(self):
        """Returns True if there is no expires time on this cookie.

        The expires time is calculated from either the max_age or
        expires attributes."""
        return self.expires_time is not None

    def is_hostonly(self):
        """Returns True if this cookie is 'host only'

        In other words, it should only be sent to the host that
        set the cookie originally."""
        return self.domain is None

    def touch(self, now=None):
        """Updates the cookie's last access time.

        now (optional)
            Time value to use. This can be in the past or the future and
            improves performance when updating multiple cookies
            simultaneously."""
        if now is None:
            now = time.time()
        self.access_time = now

    def expired(self, now=None):
        """Returns True if the cookie has expired

        now (optional)
            Time value at which to test, this can be in the past or the
            future and is largely provided to aid testing and also to
            improve performance when a large number of cookies need to
            be tested sequentially."""
        if now is None:
            now = time.time()
        if self.expires_time is None:
            return False
        else:
            return now > self.expires_time


class Section4Cookie(Cookie):

    """Represents a strict cookie definition

    The purpose of this class is wrap :class:`Cookie` to enforce
    more validation rules on the definition to ensure that the
    cookie adheres to section 4 syntax, and not just the broader
    section 5 syntax.

    Names are checked for token validity, values are checked against the
    syntax for cookie-value and the attributes are checked against the
    other constraints in the specification.

    The built-in str function will return a string that is valid against
    the section 4 syntax."""

    def __init__(self, *args, **kwargs):
        super(Section4Cookie, self).__init__(*args, **kwargs)
        # check that name is a token
        # TODO
        # check that value is a valid value
        # TODO
        # check that domain is a valid domain
        # TODO
        # check that path is a valid path
        # TODO
        # check that expires date is in UTC and is post-1600
        if self.extensions:
            attrs = set()
            if self.domain:
                attrs.add(b'domain')
            if self.path:
                attrs.add(b'path')
            if self.max_age is not None:
                attrs.add(b'max_age')
            if self.expires is not None:
                attrs.add(b'expires')
            if self.secure:
                attrs.add(b'secure')
            if self.http_only:
                attrs.add(b'httponly')
            for ext in self.extensions:
                if not ext:
                    continue
                ext = ext.split(b'=')[0].lower()
                if ext:
                    if ext in attrs:
                        raise CookieError("Duplicate attribute: %s" % ext)
                    attrs.add(ext)

    @classmethod
    def from_str(cls, src):
        """Creates a new instance from a src string

        Overridden to provide stricter parsing.  This may still appear
        more generous than expected because the strict syntax allows an
        unrestricted set of attribute extensions so unrecognised
        attributes will often be recorded but not in any useful way."""
        p = CookieParser(src)
        result = p.require_set_cookie_string(strict=True)
        p.require_end("Set-Cookie string")
        return result


class CookieDNode(object):

    def __init__(self, name):
        self.name = name
        self.public = True
        self.private = False
        self.public_depth = 0
        self.nodes = {}
        self.cookies = {}


class CookieStore(object):

    """An object that provides in-memory storage for cookies.

    There are no initialisation options.  By default, the cookie storage
    will refuse all 'domain' cookies.  That is, cookies that have a
    domain attribute.  If a domain cookie is received from a host that
    exactly matches its domain attribute then it is converted to a
    host-only cookie and *is* stored.

    This behaviour can be changed by adding exclusions (in the form of
    calls to :meth:`add_private_suffix`) or by loading in a new public
    prefix database using :meth:`set_public_list`."""

    def __init__(self):
        self.lock = threading.RLock()
        self.nodes = {}
        # effectively infinity, all domains are public we accept no
        # domain cookies at all
        self.public_depth = 255

    def set_cookie(self, url, c):
        """Store a cookie.

        urL
            A :class:`~pyslet.rfc2396.URI` instance representing the
            resource that is setting the cookie.

        c
            A :class:`Cookie` instance, typically parsed from a
            Set-Cookie header returned when requesting the resource at
            url.

        If the cookie can't be set then :class:`CookieError` is raised.
        Reasons why a cookie might be refused are a mismatch between a
        domain attribute and the url, or an attempt to set a cookie in a
        public domain, such as 'co.uk'."""
        # start with the domain of the cookie
        if not isinstance(url, params.HTTPURL) and c.http_only:
            raise CookieError(
                "Can't set http_only cookie with this protocol: %s" %
                url.scheme)
        try:
            host_str = url.host
            host = split_domain(host_str)
            host.reverse()
        except ValueError:
            # the domain is an IP address, match whole
            host = [host]
        domain = c.domain
        if domain is None:
            domain_str = host_str
            domain = host
        else:
            domain_str = domain
            domain = split_domain(domain)
            domain.reverse()
            if not domain_in_domain(domain, host):
                # you can't set this cookie from host
                raise CookieError(
                    "Can't set cookie for %s from %s" %
                    (domain_str, host_str))
        # check the path
        if c.path is None:
            if url.abs_path and url.abs_path[0] == '/':
                path = url.abs_path.split('/')
                if len(path) > 2:
                    c.path = '/'.join(path[:-1])
                else:
                    c.path = '/'
            else:
                c.path = '/'
        # now find the domain node and insert the cookie
        with self.lock:
            dnode = self
            for d in domain:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    dnode = self._expand_node(old_node, d)
            if dnode.public:
                # if this is a public node, check host
                if len(domain) == len(host):
                    # force the cookie to be host only
                    c.domain = None
                else:
                    # ignore the cookie
                    raise CookieError(
                        "Cookie for public domain %s ignored" % domain_str)
            old_cookie = dnode.cookies.get((c.name, c.path), None)
            if old_cookie is not None:
                c.creation_time = old_cookie.creation_time
                del dnode.cookies[(c.name, c.path)]
            if not c.expired():
                dnode.cookies[(c.name, c.path)] = c

    def search(self, url):
        """Searches for cookies that match a resource

        url
            A :class:`~pyslet.rfc2396.URI` instance representing the
            resource that we want to find cookies for.

        The return result is a sorted list of :class:`Cookie` objects.
        The sort order is defined in the specification, longer paths are
        sorted first, otherwise older cookies are listed before newer
        ones.

        Expired cookies are automatically removed from the repository
        and all cookies returned have their access time updated to the
        current time."""
        try:
            host_str = url.host
            host = split_domain(host_str)
            host.reverse()
        except ValueError:
            # the domain is an IP address, match whole
            host = [host]
        matched = []
        dpos = 0
        now = time.time()
        with self.lock:
            dnode = self
            for d in host:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    break
                dpos += 1
                # True if we are matching the full host
                host_match = (dpos == len(host))
                # all cookies in dnode should be considered
                expired = []
                for c in dict_values(dnode.cookies):
                    if c.expired(now):
                        expired.append(c)
                        continue
                    if c.domain is None and not host_match:
                        continue
                    # do a path match
                    if url.abs_path is None:
                        continue
                    if len(url.abs_path) < len(c.path):
                        continue
                    if not (url.abs_path.startswith(c.path) and
                            (c.path[-1] == '/' or
                             url.abs_path[len(c.path):len(c.path) + 1] ==
                             '/')):
                        continue
                    # we have a domain and path match
                    if c.secure and not isinstance(url, params.HTTPSURL):
                        continue
                    if c.http_only and not isinstance(url, params.HTTPURL):
                        continue
                    matched.append(c)
                # clean up expired cookies
                for c in expired:
                    del dnode.cookies[(c.name, c.path)]
                for c in matched:
                    c.touch(now)
        matched.sort(key=lambda x: (-len(x.path), x.creation_time))
        return matched

    def expire_cookies(self, now=None, dnode=None):
        """Expire stored cookies.

        now (optional)
            The time at which to expire the cookies, defaults to the
            current time.  This can be used to expire cookies based on
            some past or future point.

        Iterates through all stored cookies and removes any that have
        expired."""
        with self.lock:
            if dnode is None:
                dnode = self
            if now is None:
                now = time.time()
            old_node = dnode
            for dnode in dict_values(old_node.nodes):
                # all cookies in dnode should be considered
                expired = []
                for c in dict_values(dnode.cookies):
                    if c.expired(now):
                        expired.append(c)
                        continue
                # clean up expired cookies
                for c in expired:
                    del dnode.cookies[(c.name, c.path)]
                # now recurse
                self.expire_cookies(now, dnode)

    def end_session(self, now=None, dnode=None):
        """Expire all session cookies.

        now (optional)
            The time at which to expire cookies.  See
            :meth:`expire_cookies` for details.

        Iterates through all stored cookies and removes any session
        cookies *in addition*  to any that have expired."""
        with self.lock:
            if dnode is None:
                dnode = self
            if now is None:
                now = time.time()
            old_node = dnode
            for dnode in dict_values(old_node.nodes):
                # all cookies in dnode should be considered
                expired = []
                for c in dict_values(dnode.cookies):
                    if c.expired(now) or not c.is_persistent():
                        expired.append(c)
                        continue
                # clean up expired cookies
                for c in expired:
                    del dnode.cookies[(c.name, c.path)]
                # now recurse
                self.end_session(now, dnode)

    def add_public_suffix(self, suffix):
        """Marks a domain suffix as being public.

        suffix
            A string: a public suffix, may contain wild-card
            characters to match any entire label, for example:
            "*.uk", "*.tokyo.jp", "com"

        Once a domain suffix is marked as being public *future* cookies
        will not be stored against that suffix (except in the unusual
        case where a cookie is 'host only' and the host name is a public
        suffix)."""
        domain = split_domain(suffix, allow_wildcard=True)
        domain.reverse()
        with self.lock:
            dnode = self
            dcount = len(domain)
            for d in domain:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    if d == '*':
                        # wild card handling
                        if dcount == 1:
                            # rule finished with a wild card
                            if old_node.public_depth < 1:
                                old_node.public_depth = 1
                            # mark everthing at this level public
                            for child in dict_values(old_node.nodes):
                                if not child.private:
                                    child.public = True
                        else:
                            # create a wild node
                            dnode = self._expand_node(old_node, d)
                    else:
                        # create a new node
                        dnode = self._expand_node(old_node, d)
                        # force this to be public
                        dnode.public = True
                elif dcount == 1:
                    # this is the last label in the suffix, mark it
                    # as public if it isn't explicitly private
                    if not dnode.private:
                        dnode.public = True
                    if d == '*':
                        # mark all children public too
                        for child in dict_values(old_node.nodes):
                            if not child.private:
                                child.public = True
                dcount -= 1

    def add_private_suffix(self, suffix):
        """Marks a domain suffix as being private.

        suffix
            A string: a public suffix, may contain wild-card
            characters to match any entire label, for example:
            "example.co.uk", "\\*.tokyo.jp", "com"

        This method is required to override an existing public rule,
        thereby ensuring that *future* cookies can be stored against
        domains matching this suffix."""
        domain = split_domain(suffix, allow_wildcard=True)
        domain.reverse()
        with self.lock:
            dnode = self
            dcount = len(domain)
            for d in domain:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    if d == '*':
                        # wild card handling
                        if dcount == 1:
                            # rule finished with a wild card - this is
                            # supposed to be an exception to a previous wild
                            # card rule so assume all domains here are
                            # excepted and hence private
                            old_node.public_depth = 0
                            # mark everthing at this level private
                            for child in dict_values(old_node.nodes):
                                child.private = True
                                child.public = False
                        else:
                            # create a wild node
                            dnode = self._expand_node(old_node, d)
                    else:
                        # create a new node
                        dnode = self._expand_node(old_node, d)
                        if dcount == 1:
                            # private node
                            dnode.private = True
                            dnode.public = False
                elif dcount == 1:
                    # this is the last label in the suffix, mark it
                    # as private
                    dnode.private = True
                    dnode.public = False
                    if d == '*':
                        # mark all children public too
                        for child in dict_values(old_node.nodes):
                            child.private = True
                            child.public = False
                dcount -= 1

    @classmethod
    def fetch_public_suffix_list(
            cls, fpath,
            src="https://publicsuffix.org/list/effective_tld_names.dat",
            overwrite=False):
        """Fetches the public suffix list and saves to fpath

        fpath
            A local file path to save the file in

        src
            A string or :class:`~pyslet.rfc2396.URI` instance pointing
            at the file to retrieve.  It default to the data file
            https://publicsuffix.org/list/effective_tld_names.dat

        overwrite (Default: False)
            A flag to force an overwrite of an existing file at fpath,
            by default, if fpath already exists this method returns
            without doing anything."""
        from . import client as http
        if not os.path.exists(fpath):
            c = http.Client()
            request = http.ClientRequest(src)
            c.process_request(request)
            if request.status == 200 and request.res_body:
                with open(fpath, 'wb') as f:
                    f.write(request.res_body)
            else:
                logging.error("Failed to download list from %s", str(src))
            c.close()

    def set_public_list(self, black_list, tld_depth=1):
        """Loads a new public suffix list

        black_list
            A string containing a list of public suffixes in the format
            defined by: https://publicsuffix.org/list/

        tld_depth (Default: 1)
            The depth of domain that will be automatically treated as
            public.  The default is 1, meaning that all top-level domains
            will be treated as public.

        This methods loads data from a public list using calls to
        :meth:`add_public_prefix` and :meth:`add_private_prefix`, the
        latter being for exclusion rules.

        If you use the full list published by the Public Suffix List
        project it is safe to use the default tld_depth value of 1:

        https://publicsuffix.org/list/effective_tld_names.dat

        If you want to load a much smaller list then you should focus
        on a large value for tld_depth (255 for example) and documenting
        exclusions only.  For example::

            // Exclusion list
            // Accept domain cookies for example.com, example.co.uk
            !example.com
            !example.co.uk"""
        self.public_depth = tld_depth
        lines = black_list.splitlines()
        for line in lines:
            if not line or line[0:2] == '//':
                continue
            line = line.split()
            line = line[0]
            if not line:
                continue
            try:
                if line[0:1] == '!':
                    self.add_private_suffix(line[1:])
                else:
                    self.add_public_suffix(line)
            except ValueError:
                logging.warning("Ignoring bad rule in black_list: %s",
                                line)

    def test_public_domain(self, domain_str):
        """Test if a domain is public

        domain_str
            A domain string, e.g., "www.example.com"

        Returns True if this domain is marked as public, False
        otherwise."""
        domain = split_domain(domain_str)
        domain.reverse()
        with self.lock:
            dnode = self
            dcount = len(domain)
            for d in domain:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    # dcount is the number of labels left in domain
                    if dcount > old_node.public_depth:
                        return False
                    else:
                        return True
                elif dcount == 1:
                    # last label, this domain matches this rule exactly
                    return dnode.public
                dcount -= 1

    def get_registered_domain(self, domain_str, u_labels=False):
        """Returns the publicly registered portion of a domain

        domain_str
            A domain string, e.g., "www.example.com"

        u_labels (Default: False)
            Flag indicating whether or not to return unicode
            labels instead of encoded ASCII Labels.

        Compares this domain against the database of public domains and
        returns the publicly registered part of the domain.  For
        example, www.example.com would typically return example.com and
        www.example.co.uk would typically return example.co.uk.

        If domain_str is already a publicly registered domain then it
        returns None. If domain_str is itself None, None is also
        returned.

        Initially, all domains are marked as public so this function
        will always return None.  It iss intended for use after a public
        list has been loaded, such as the public suffix list (see
        :meth:`set_public_list`)."""
        if domain_str is None:
            return None
        domain = split_domain(domain_str)
        domain.reverse()
        if domain and domain[-1] == '':
            # ignore domains with a leading '.'
            return None
        with self.lock:
            dnode = self
            dcount = len(domain)
            matched = []
            for d in domain:
                old_node = dnode
                dnode = old_node.nodes.get(d, None)
                if dnode is None:
                    # dcount is the number of labels left in domain
                    if dcount > old_node.public_depth:
                        # private domain with public prefix
                        matched = matched + domain[
                            -dcount:len(domain) - dcount +
                            old_node.public_depth + 1]
                    else:
                        # public domain
                        matched = None
                    break
                elif not dnode.public:
                    # that's it, don't add any more labels after this one
                    matched.append(d)
                    break
                elif dcount == 1:
                    # last label, this domain matches this rule exactly
                    matched = None
                    break
                matched.append(d)
                dcount -= 1
        if matched is None:
            return None
        matched.reverse()
        if u_labels:
            matched = [u[4:].decode('punycode') if is_a_label(u) else
                       u.decode('utf-8') for u in matched]
        return '.'.join(matched)

    def check_public_suffix(self, domain_str, match_str):
        """See Public Suffix Test Data for details.

        http://mxr.mozilla.org/mozilla-central/source/netwerk/test/unit/data/test_psl.txt?raw=1

        Returns True if there is a match, False otherwise. Negative
        results are logged at ERROR level.  Used for testing the public
        suffixes loaded with :meth:`set_public_list`."""
        matched = self.get_registered_domain(
            domain_str, is_unicode(match_str))
        if matched is None:
            if match_str is None:
                return True
            else:
                logging.error("Bad match: %s and %s; found None",
                              domain_str, match_str)
                return False
        elif match_str != matched:
            logging.error("Bad match: %s and %s; found %s",
                          domain_str, match_str, repr(matched))
            return False
        else:
            return True

    def _copy_node(self, src_node, label=None):
        if label is None:
            label = src_node.name
        dnode = CookieDNode(label)
        dnode.public = src_node.public
        dnode.private = src_node.private
        dnode.public_depth = src_node.public_depth
        for src_child in dict_values(src_node.nodes):
            child = self._copy_node(src_child)
            dnode.nodes[child.name] = child
        return dnode

    def _expand_node(self, from_node, label):
        # create a new node
        if label != '*' and '*' in from_node.nodes:
            # expand by copying the wild node
            wild_node = from_node.nodes['*']
            # copies any children too, feels painful but there won't
            # often be children and when then is there won't be many
            dnode = self._copy_node(wild_node, label)
        else:
            dnode = CookieDNode(label)
            # expand by copying information from parent
            if from_node.public_depth:
                dnode.public = True
                dnode.public_depth = from_node.public_depth - 1
            else:
                dnode.public = False
                dnode.public_depth = 0
        from_node.nodes[label] = dnode
        return dnode



#! /usr/bin/env python


from . import core
from .. import html401 as html
from ..py2 import is_text
from ..xml import structures as xml


class BodyElement(core.QTIElement):

    """The root class of all content objects in the item content model is the
    bodyElement. It defines a number of attributes that are common to all
    elements of the content model::

            <xsd:attributeGroup name="bodyElement.AttrGroup">
                    <xsd:attribute name="id" type="identifier.Type"
                    use="optional"/>
                    <xsd:attribute name="class" use="optional">
                            <xsd:simpleType>
                                    <xsd:list itemType="styleclass.Type"/>
                            </xsd:simpleType>
                    </xsd:attribute>
                    <xsd:attribute ref="xml:lang"/>
                    <xsd:attribute name="label" type="string256.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLATTR_id = ('id', core.ValidateIdentifier, lambda x: x)
    XMLATTR_class = ('style_class', None, None, list)
    XMLATTR_label = 'label'

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.id = None
        self.label = None

    def render_html(self, parent, profile, item_state):
        """Renders this element in html form, adding nodes to *parent*.  This
        method effectively overrides
        :py:class:`html401.XHTMLElement.render_html` enabling QTI and
        XHTML elements to be mixed freely.

        The state of the item (e.g., the values of any controls), is taken from
        *item_state*, a :py:class:`variables.ItemSessionState` instance."""
        raise NotImplementedError(self.__class__.__name__ + ".render_html")

    def render_html_children(self, parent, profile, item_state):
        """Renders this element's children to an external document represented
        by the *parent* node"""
        for child in self.get_children():
            if is_text(child):
                parent.add_data(child)
            else:
                child.render_html(parent, profile, item_state)


TextElements = {
    'abbr': ('id', 'class', 'label'),
    'acronym': ('id', 'class', 'label'),
    'address': ('id', 'class', 'label'),
    'blockquote': ('id', 'class', 'label'),
    'br': ('id', 'class', 'label'),
    'cite': ('id', 'class', 'label'),
    'code': ('id', 'class', 'label'),
    'dfn': ('id', 'class', 'label'),
    'div': ('id', 'class', 'label'),
    'em': ('id', 'class', 'label'),
    'h1': ('id', 'class', 'label'),
    'h2': ('id', 'class', 'label'),
    'h3': ('id', 'class', 'label'),
    'h4': ('id', 'class', 'label'),
    'h5': ('id', 'class', 'label'),
    'h6': ('id', 'class', 'label'),
    'kbd': ('id', 'class', 'label'),
    'p': ('id', 'class', 'label'),
    'pre': ('id', 'class', 'label'),
    'q': ('id', 'class', 'label'),
    'samp': ('id', 'class', 'label'),
    'span': ('id', 'class', 'label'),
    'strong': ('id', 'class', 'label'),
    'var': ('id', 'class', 'label')
}       #: Basic text formatting elements

ListElements = {
    'dl': ('id', 'class', 'label'),
    'dt': ('id', 'class', 'label'),
    'dd': ('id', 'class', 'label'),
    'ol': ('id', 'class', 'label'),
    'ul': ('id', 'class', 'label'),
    'li': ('id', 'class', 'label')
}       #: Elements required for lists


ObjectElements = {
    'object': ('id', 'class', 'label', 'data', 'type', 'width', 'height'),
    'param': ('id', 'class', 'label', 'name', 'value', 'valuetype', 'type')
}       #: The object element


PresentationElements = {
    'caption': ('id', 'class', 'label'),
    'col': ('id', 'class', 'label', 'span'),
    'colgroup': ('id', 'class', 'label', 'span'),
    'table': ('id', 'class', 'label', 'summary'),
    'tbody': ('id', 'class', 'label'),
    'tfoot': ('id', 'class', 'label'),
    'thead': ('id', 'class', 'label'),
    'tr': ('id', 'class', 'label'),
    'th': ('id', 'class', 'label', 'headers', 'scope', 'abbr', 'axis',
           'rowspan', 'colspan'),
    'td': ('id', 'class', 'label', 'headers', 'scope', 'abbr', 'axis',
           'rowspan', 'colspan')
}  #: Tables

ImageElement = {
    'img': ('id', 'class', 'label', 'src', 'alt', 'longdesc', 'height',
            'width')}  #: Images


HypertextElement = {
    'a': ('id', 'class', 'label', 'href', 'type')
}       #: Hyperlinks


HTMLProfile = {}        #: The full HTML profile defined by QTI

HTMLProfile.update(TextElements)
HTMLProfile.update(ListElements)
HTMLProfile.update(ObjectElements)
HTMLProfile.update(PresentationElements)
HTMLProfile.update(ImageElement)
HTMLProfile.update(HypertextElement)


def fix_html_namespace(e):
    """Fixes e and all children to be in the QTI namespace"""
    if e.ns == html.XHTML_NAMESPACE:
        name = e.xmlname.lower()
        if name in core.QTI_HTML_PROFILE:
            e.set_xmlname((core.IMSQTI_NAMESPACE, name))
    for e in e.get_children():
        if is_text(e):
            continue
        fix_html_namespace(e)


class ItemBody(BodyElement):

    """The item body contains the text, graphics, media objects, and
    interactions that describe the item's content and information about how it
    is structured::

            <xsd:attributeGroup name="itemBody.AttrGroup">
                    <xsd:attributeGroup ref="bodyElement.AttrGroup"/>
            </xsd:attributeGroup>

            <xsd:group name="itemBody.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="block.ElementGroup" minOccurs="0"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'itemBody')
    XMLCONTENT = xml.ElementContent

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.BlockMixin):
            return BodyElement.add_child(self, child_class, name)
        else:
            raise core.QTIValidityError(
                "%s (%s) in %s" %
                (repr(name),
                 child_class.__name__,
                 self.__class__.__name__))

    def render_html(self, parent, profile, item_state):
        """Overrides :py:meth:`BodyElement.render_html`, the result is always a
        Div with class set to "itemBody".  Unlike other such method *parent*
        may by None, in which case a new parentless Div is created."""
        if parent:
            html_div = parent.add_child(html.Div)
        else:
            html_div = html.Div(None)
        html_div.style_class = ["itemBody"]
        self.render_html_children(html_div, profile, item_state)


class FlowContainerMixin(object):

    """Mixin class used for objects that can contain flows."""

    def pretty_print(self):
        """True if this object should be pretty printed.

        This is similar to the algorithm we use in HTML flow containers,
        suppressing pretty printing if we have inline elements (ignoring
        non-trivial data).  This could be refactored in future."""
        for child in self.get_children():
            if is_text(child):
                for c in child:
                    if not xml.is_s(c):
                        return False
            elif isinstance(child, html.InlineMixin):
                return False
        return True


class RubricBlock(html.BlockMixin, BodyElement):

    """Represent the rubricBlock element.

    <xsd:attributeGroup name="rubricBlock.AttrGroup">
            <xsd:attributeGroup ref="simpleBlock.AttrGroup"/>
            <xsd:attribute name="view" use="required">
                    <xsd:simpleType>
                            <xsd:list itemType="view.Type"/>
                    </xsd:simpleType>
            </xsd:attribute>
    </xsd:attributeGroup>

    <xsd:group name="rubricBlock.ContentGroup">
            <xsd:sequence>
                    <xsd:group ref="simpleBlock.ContentGroup"/>
            </xsd:sequence>
    </xsd:group>
    """
    XMLNAME = (core.IMSQTI_NAMESPACE, 'rubricBlock')
    XMLATTR_view = (
        'view', core.View.from_str_lower, core.View.to_str, dict)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        BodyElement.__init__(self, parent)
        self.view = {}

    def add_view(self, view):
        if is_text(view):
            view = core.View.from_str_lower(view.strip())
        view_value = core.View.to_str(view)
        if view_value:
            self.view[view] = view_value
        else:
            raise ValueError("illegal value for view: %s" % view)

    # need to constrain content to html.BlockMixin
    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.BlockMixin):
            return BodyElement.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))



#! /usr/bin/env python

from .. import html401 as html
from ..pep8 import (
    MigratedClass,
    old_function,
    old_method)
from ..xml import structures as xml
from ..xml import namespace as xmlns
from ..xml import xsdatatypes as xsi


IMSQTI_NAMESPACE = "http://www.imsglobal.org/xsd/imsqti_v2p1"
"""The namespace used to recognise elements in XML documents."""

IMSQTI_SCHEMALOCATION = "http://www.imsglobal.org/xsd/imsqti_v2p1.xsd"
"""The location of the QTI 2.1 schema file on the IMS website."""

IMSQTI_ITEM_RESOURCETYPE = "imsqti_item_xmlv2p1"
"""The resource type to use for the QTI 2.1 items when added to content
packages."""

QTI_HTML_PROFILE = set((
    'abbr', 'acronym', 'address', 'blockquote', 'br', 'cite', 'code', 'dfn',
    'div', 'em', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'kbd', 'p', 'pre', 'q',
    'samp', 'span', 'strong', 'var', 'dl', 'dt', 'dd', 'ol', 'ul', 'li',
    'object', 'param', 'b', 'big', 'hr', 'i', 'small', 'sub', 'sup', 'tt',
    'caption', 'col', 'colgroup', 'table', 'tbody', 'tfoot', 'thead', 'td',
    'th', 'tr', 'img', 'a'))


class QTIError(Exception):

    """Abstract class used for all QTI v2 exceptions."""
    pass


class DeclarationError(QTIError):

    """Error raised when a variable declaration is invalid."""
    pass


class ProcessingError(QTIError):

    """Error raised when an invalid processing element is encountered."""
    pass


class SelectionError(QTIError):

    """Error raised when there is a problem with creating test forms."""
    pass


class QTIValidityError(QTIError):
    pass


@old_function('ValidateIdentifier')
def validate_identifier(value, prefix='_'):
    """Decodes an identifier from a string::

        <xsd:simpleType name="identifier.Type">
                <xsd:restriction base="xsd:NCName"/>
        </xsd:simpleType>

    This function takes a string that is supposed to match the
    production for NCName in XML and forces it to comply by replacing
    illegal characters with '_', except the ':' which is replaced with a
    hyphen for compatibility with previous versions of the QTI migration
    script.  If name starts with a valid name character but not a valid
    name start character, it is prefixed with '_' too, but the prefix
    string used can be overridden."""
    if value:
        good_name = []
        if not xml.is_name_start_char(value[0]):
            good_name.append(prefix)
        elif value[0] == ':':
            # Previous versions of the migrate script didn't catch this problem
            # as a result, we deviate from its broken behaviour of using '-'
            # by using the prefix too.
            good_name.append(prefix)
        for c in value:
            if c == ':':
                good_name.append('-')
            elif xml.is_name_char(c):
                good_name.append(c)
            else:
                good_name.append('_')
        return ''.join(good_name)
    else:
        return prefix


class Orientation(xsi.Enumeration):

    """Orientation attribute values provide a hint to rendering systems that an
    element has an inherent vertical or horizontal interpretation::

            <xsd:simpleType name="orientation.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="horizontal"/>
                            <xsd:enumeration value="vertical"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above orientations.  Usage example::

            Orientation.horizontal

    Note that::

            Orientation.DEFAULT == None

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'horizontal': 1,
        'vertical': 2
    }


class Shape(xsi.Enumeration):

    """A value of a shape is always accompanied by coordinates and an associated
    image which provides a context for interpreting them::

            <xsd:simpleType name="shape.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="circle"/>
                            <xsd:enumeration value="default"/>
                            <xsd:enumeration value="ellipse"/>
                            <xsd:enumeration value="poly"/>
                            <xsd:enumeration value="rect"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above types of Shape.  Usage example::

            Shape.circle

    Note that::

            Shape.DEFAULT == Shape.default

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'circle': 1,
        'default': 2,
        'ellipse': 3,
        'poly': 4,
        'rect': 5
    }
    aliases = {
        None: 'default'
    }


def calculate_shape_bounds(shape, coords):
    """Calculates a bounding rectangle from a Shape value and a
    :py:class:`pyslet.html401.Coords` instance."""
    if shape == Shape.circle:
        return [coords[0].resolve_value(1024) - coords[2].resolve_value(768),
                coords[1].resolve_value(768) - coords[2].resolve_value(768),
                coords[0].resolve_value(1024) + coords[2].resolve_value(768),
                coords[1].resolve_value(768) + coords[2].resolve_value(768)]
    elif shape == Shape.default:
        return [0, 0, 1024, 768]
    elif shape == Shape.ellipse:
        return [coords[0].resolve_value(1024) - coords[2].resolve_value(1024),
                coords[1].resolve_value(768) - coords[3].resolve_value(768),
                coords[0].resolve_value(1024) + coords[2].resolve_value(1024),
                coords[1].resolve_value(768) + coords[3].resolve_value(768)]
    elif shape == Shape.poly:
        output = [coords[0].resolve_value(1024), coords[1].resolve_value(768),
                  coords[0].resolve_value(1024), coords[1].resolve_value(768)]
        i = 1
        while 2 * i + 1 < len(coords):
            x = coords[2 * i].resolve_value(1024)
            y = coords[2 * i + 1].resolve_value(768)
            if x < output[0]:
                output[0] = x
            elif x > output[2]:
                output[2] = x
            if y < output[1]:
                output[1] = y
            elif y > output[3]:
                output[3] = y
        return output
    elif shape == Shape.rect:
        return [coords[0].resolve_value(1024), coords[1].resolve_value(768),
                coords[2].resolve_value(1024), coords[3].resolve_value(768)]
    else:
        raise ValueError("Unknown value for shape: %s" % str(shape))


def offset_shape(shape, coords, xoffset, yoffset):
    """Interprets the shape and coords relative to the given offset and maps
    them back to the origin.

    In other words, xoffset and yoffset are subtracted from the coordinates."""
    if shape == Shape.circle:
        coords[0].add(-xoffset)
        coords[1].add(-yoffset)
    elif shape == Shape.default:
        pass
    elif shape == Shape.ellipse:
        coords[0].add(-xoffset)
        coords[1].add(-yoffset)
    elif shape == Shape.poly:
        i = 0
        while 2 * i + 1 < len(coords):
            coords[2 * i].add(-xoffset)
            coords[2 * i + 1].add(-yoffset)
    elif shape == Shape.rect:
        coords[0].add(-xoffset)
        coords[1].add(-yoffset)
        coords[2].add(-xoffset)
        coords[3].add(-yoffset)
    else:
        raise ValueError("Unknown value for shape: %s" % str(shape))


class ShapeElementMixin(MigratedClass):
    XMLATTR_shape = ('shape', Shape.from_str_lower, Shape.to_str)
    XMLATTR_coords = ('coords', html.Coords.from_str, html.Coords.__unicode__)

    def __init__(self):
        self.shape = Shape.DEFAULT  # : The shape
        self.coords = None			#: A list of Length values

    @old_method('TestPoint')
    def test_point(self, point, width, height):
        """Tests *point* to see if it is in this area."""
        x, y = point
        if self.shape == Shape.circle:
            return self.coords.test_circle(x, y, width, height)
        elif self.shape == Shape.default:
            # The entire region
            return x >= 0 and y >= 0 and (
                width is None or x <= width) and (
                height is None or y <= height)
        elif self.shape == Shape.ellipse:
            # Ellipse is deprecated because there is no HTML equivalent test
            return self.TestEllipse(x, y, width, height)
        elif self.shape == Shape.poly:
            return self.coords.test_poly(x, y, width, height)
        elif self.shape == Shape.rect:
            return self.coords.test_rect(x, y, width, height)
        else:
            raise ValueError("Unknown Shape type")

    @old_method('TestEllipse')
    def test_ellipse(self, x, y, width, height):
        """Tests an x,y point against an ellipse with these coordinates.

        HTML does not define ellipse, we take our definition from the QTI
        specification itself: center-x, center-y, x-radius, y-radius."""
        if len(self.coords.values) < 4:
            raise ValueError(
                "Ellipse test requires 4 coordinates: %s" % str(
                    self.coords.values))
        dx = x - self.coords.values[0].resolve_value(width)
        dy = y - self.coords.values[1].resolve_value(height)
        rx = self.coords.values[2].resolve_value(width)
        ry = self.coords.values[3].resolve_value(height)
        return dx * dx * ry * ry + dy * dy * rx * rx <= rx * rx * ry * ry


class ShowHide(xsi.Enumeration):

    """Used to control content visibility with variables
    ::

            <xsd:simpleType name="showHide.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="hide"/>
                            <xsd:enumeration value="show"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Note that ShowHide.DEFAULT == ShowHide.show"""
    decode = {
        'show': 1,
        'hide': 2
    }
    aliases = {
        None: 'show'
    }


class View(xsi.EnumerationNoCase):

    """Used to represent roles when restricting view::

            <xsd:simpleType name="view.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="author"/>
                            <xsd:enumeration value="candidate"/>
                            <xsd:enumeration value="proctor"/>
                            <xsd:enumeration value="scorer"/>
                            <xsd:enumeration value="testConstructor"/>
                            <xsd:enumeration value="tutor"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above views.  Usage example::

            View.candidate

    There is no default view.  Views are represented in XML as space-separated
    lists of values.  Typical usage::

            view=View.DecodeValueDict("tutor scorer")
            # returns...
            { View.tutor:'tutor', View.scorer:'scorer' }
            View.EncodeValueDict(view)
            # returns...
            "scorer tutor"

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'author': 1,
        'candidate': 2,
        'proctor': 3,
        'scorer': 4,
        'testConstructor': 5,
        'tutor': 6
    }


class QTIElement(xmlns.XMLNSElement):

    """Basic element to represent all QTI elements"""

    def add_to_cpresource(self, cp, resource, been_there):
        """We need to add any files with URL's in the local file system to the
        content package.

        been_there is a dictionary we use for mapping URLs to File objects so
        that we don't keep adding the same linked resource multiple times.

        This implementation is a little more horrid, we avoid circular module
        references by playing dumb about our children.  HTML doesn't actually
        know anything about QTI even though QTI wants to define children for
        some XHTML elements so we pass the call only to "CP-Aware" elements."""
        for child in self.get_children():
            if hasattr(child, 'add_to_cpresource'):
                child.add_to_cpresource(cp, resource, been_there)


class DeclarationContainer(MigratedClass):

    """An abstract mix-in class used to manage a dictionary of variable
    declarations."""

    def __init__(self):
        #: a dictionary of outcome variable declarations
        self.declarations = {}

    @old_method('RegisterDeclaration')
    def register_declaration(self, declaration):
        if declaration.identifier in self.declarations:
            raise DeclarationError
        else:
            self.declarations[declaration.identifier] = declaration

    @old_method('IsDeclared')
    def is_declared(self, identifier):
        return identifier in self.declarations

    @old_method('GetDeclaration')
    def get_declaration(self, identifier):
        return self.declarations.get(identifier, None)


@old_function('GetTemplateRef')
def get_template_ref(value):
    """Given a string used to set an attribute of an *...orTemplateRef* type
    this function returns the name of the variable being referred to or None if
    the value does not look like a template variable reference."""
    if value.startswith('{') and value.endswith('}'):
        id_value = value[1:-1]
        if xsi.is_valid_ncname(id_value):
            return id_value
    return None
